function [D, X] = dl_gauss_l1(Y, D, sigmas, lam, gam, niter, nit_fista, step_fista)
% Input:
%    Y          - matrix of training signals (on columns)
%    D          - dictionary, central atoms
%    sigmas     - vector of standard deviations associated with the central atoms
%    lam        - weight of the error term
%    gam        - weight of the 1-norm term
%    niter      - number of iterations
%    nit_fista  - number of iterations for FISTA
%    step_fista - step size in FISTA
% Output:
%    D          - learned dictionary
%    X          - representation matrix

% BD 25.08.2023

[m,n] = size(D);
N = size(Y,2);
if length(sigmas) == 1
  sigmas = sigmas * ones(n,1);
else
  sigmas = sigmas(:);
end
s2 = 1 ./ sigmas ./ sigmas;

X = zeros(n,N);

for it = 1 : niter
  Dnew = zeros(m,n);

  for iN = 1 : N
    y = Y(:,iN);
    x = X(:,iN);
    x = my_fista(y, D, gam/lam, step_fista, x, nit_fista);

    r = y - D*x;     % residual
    Da = D;

    % update actual atoms one by one, in random order
    p = randperm(n);
    for j = p
      r = r + Da(:,j)*x(j);  % remove contribution of current atom
      % update atom and normalize
      Da(:,j) = (1/(lam*x(j)*x(j) + s2(j))) * (lam*x(j)*r + s2(j)*D(:,j));
      % update representation coefficient
      x(j) = wthresh(Da(:,j)'*r, 's', gam/lam/2);
      % update residual
      r = r - Da(:,j)*x(j);
    end
    Da = normc(Da); % atom normalization
    X(:,iN) = x;

    support = find(x~=0);
    Dnew(:,support) = Dnew(:,support) + Da(:,support);
  end

  Dnew = normc(Dnew);
  % norm(D-Dnew);
  D = Dnew;
end

%---------------------------------------------
function x = my_fista(y, D, lam, t, x, nit)

x1 = x;
s1 = 1;
for it = 1 : nit
  x2 = x1;
  s2 = s1;
  x1 = wthresh(x - 2*t*D'*(D*x - y), 's', t*lam);
  s1 = (1 + sqrt(1+4*s2*s2))/2;
  x = x1 + ((s2-1)/s1)*(x1-x2);
end
