function [x, Da, err, proba, obj] = gauss_l1_s(y, D, sigmas, lam, gam, niter, nit_fista, step_fista)
% Compute sparse representation for dictionary with Gaussian atoms,
% simpler algorithm, using directly soft thresholding for updating the coefficients.
% FISTA is used only for initialization.
% Input:
%    y          - signal to be represented
%    D          - dictionary, central atoms
%    sigmas     - vector of standard deviations associated with the central atoms
%    lam        - weight of the error term
%    gam        - weight of the 1-norm term
%    niter      - number of iterations
%    nit_fista  - number of iterations for FISTA
%    step_fista - step size in FISTA
% Output:
%    x          - representation vector
%    Da         - dictionary, actual atoms
%    err        - representation errors (vector of length niter) 
%    proba      - "probabilities" associated with the representation (or a
%                 measure of the distance to the actual atoms)
%    obj        - values of the whole objective function

% BD 25.08.2023

[~,n] = size(D);
if length(sigmas) == 1
  sigmas = sigmas * ones(n,1);
end
s2 = 1 ./ sigmas ./ sigmas;

% initialization
Da = D;     % actual atoms are the central ones
x = zeros(n,1);
x = my_fista(y, Da, gam/lam, step_fista, x, nit_fista);
obj = zeros(niter,1);

for it = 1 : niter
  r = y - Da*x;     % residual

  % update actual atoms one by one, in random order
  p = randperm(n);
  for j = p
    r = r + Da(:,j)*x(j);  % remove contribution of current atom
    % update atom and normalize
    Da(:,j) = (1/(lam*x(j)*x(j) + s2(j))) * (lam*x(j)*r + s2(j)*D(:,j));
    % update representation coefficient
    x(j) = wthresh(Da(:,j)'*r, 's', gam/lam/2);
    % update residual
    r = r - Da(:,j)*x(j);
  end
  Da = normc(Da); % atom normalization

  % compute objective output values
  err(it) = norm(y-Da*x);
  proba(it) = norm((Da-D)*diag(1/sigmas/2), 'fro')^2;
  obj(it) = norm((Da-D)*diag(1/sigmas), 'fro')^2 + lam*norm(y-Da*x)^2 + gam*norm(x,1);
end


%---------------------------------------------
function x = my_fista(y, D, lam, t, x, nit)

x1 = x;
s1 = 1;
for it = 1 : nit
  x2 = x1;
  s2 = s1;
  x1 = wthresh(x - 2*t*D'*(D*x - y), 's', t*lam);
  s1 = (1 + sqrt(1+4*s2*s2))/2;
  x = x1 + ((s2-1)/s1)*(x1-x2);
end
