% Main test program for Gaussian DL algorithms on ADBench datasets.
% Programs for paper "Sparse representations with Gaussian atoms",
% by D.C. Ilie-Ablachim, A. Baltoiu, B.Dumitrescu.
% The algorithms called Gauss-L1 and DL-Gauss-L1 in the paper are tested here.

% BD 23.08.2023

%% ----------------------------------------
data_file_path = '../DATA_DEPENDENCY/dependency_'; % where the data are
data_file_names = {'annthyroid', 'breastw', 'cardio', 'cover', 'fault', ...
                   'glass', 'Hepatitis', 'Ionosphere', 'landsat', 'letter', ...
                   'Lymphography', 'magic.gamma', 'mammography', 'musk', 'pendigits', ...
                   'Pima', 'satellite', 'satimage', 'SpamBase', 'Stamps', ...
                   'thyroid', 'vertebral', 'vowels', 'Waveform', 'WBC', ...
                   'WDBC', 'Wilt', 'wine', 'WPBC', 'yeast' }; % full file list
data_file_names = {'wine', 'yeast'}; % this line is only for a first quick test
results_file_name = ['res_gauss_', int2str(randi(9999)), '.csv'];  % the results file name has a random part

n_tests = 5; % number of repetitions
train_data_ratio = 0.7;

overcompleteness_list = [2]; % overcompleteness factors of the dictionary (may be a vector)
lam_list = [1];  % list of lambda factors (may be a vector)
gam_list = 1 * lam_list;  % there should be some proportionality here (may have different values)

sigma_max = 0.1;      % max and min standard deviations of the atoms
sigma_min = 0.01;
nit_gauss_repr = 100; % number of iterations for Gauss-L1
nit_gauss_dl = 100;   % number of iterations for DL-Gauss-L1
nit_dl = 40;          % number of iterations for AK-SVD
s_dl = 2;             % sparsity level for AK-SVD
nit_fista = 50;       % number of FISTA iterations...
step_fista = 0.0001;  % step in FISTA

method_names = {'AK-SVD + Gauss-L1', 'DL-Gauss-L1'};

% ----------------------------------------
n_datasets = length(data_file_names);
n_methods = length(method_names);
n_params = length(overcompleteness_list)*length(lam_list);

% prepare file for results; write header
fid_res = fopen(results_file_name,'w');
fprintf(fid_res, "%s,%s,", 'Dataset', 'Method');
for i_over = 1:length(overcompleteness_list)
  for i_s = 1 : length(lam_list)
    fprintf(fid_res, "%s,", ['o=',num2str(overcompleteness_list(i_over)), ' lam=', int2str(lam_list(i_s)), ' [err]']);
    fprintf(fid_res, "%s,", ['o=',num2str(overcompleteness_list(i_over)), ' lam=', int2str(lam_list(i_s)), ' [proba]']);
    fprintf(fid_res, "%s,", ['o=',num2str(overcompleteness_list(i_over)), ' lam=', int2str(lam_list(i_s)), ' [obj]']);
  end
end
fprintf(fid_res, "\n");

for i=1:n_methods
  res_methods_err{i} = zeros(n_datasets,n_params);
  res_methods_proba{i} = zeros(n_datasets,n_params);
  res_methods_obj{i} = zeros(n_datasets,n_params);
  time_methods_train{i} = zeros(n_datasets,n_params);
  time_methods_test{i} = zeros(n_datasets,n_params);
end

for id = 1 : n_datasets
  fprintf("%s\n", data_file_names{id})
  filename = [data_file_path, data_file_names{id}];
  load(filename)
  X = double(Y);   % data are in X, labels are in y
  y = labels;
  clear Y, clear labels
  X = X';          % signals are on columns
  [m,N] = size(X);

  % preprocessing: Z-score normalization
  vm = mean(X,2);
  X = double(X) - repmat(vm,1,N);         % subtract mean
  sm = std(X,0,2);
  X = X ./ repmat(sm,1,N);
  
  col_normal = find(y==0);
  col_anomaly = find(y~=0);
  Nanomaly = length(col_anomaly);
  Nnormal = N - Nanomaly;
  anomaly_ratio = Nanomaly/N;
  
  Ntrain = round(train_data_ratio*N);
  Ntest = N - Ntrain;
  N_anomaly_train = round(anomaly_ratio*Ntrain);
  N_normal_train = Ntrain - N_anomaly_train;
  N_anomaly_test = Nanomaly - N_anomaly_train;
  N_normal_test = Nnormal - N_normal_train;
  
  for im = 1:n_methods
    res_roc_err{im} = zeros(n_tests, n_params);
    res_roc_proba{im} = zeros(n_tests, n_params);
    res_roc_obj{im} = zeros(n_tests, n_params);
    res_err{im} = zeros(n_tests, n_params);
    res_proba{im} =  zeros(n_tests, n_params);
    res_obj{im} = zeros(n_tests, n_params);
    time_train{im} = zeros(n_tests, n_params);
    time_test{im} = zeros(n_tests, n_params);
  end
  
  for it = 1 : n_tests
    % split train and test data, respecting proportion of anomaly
    p1 = randperm(Nnormal);
    p2 = randperm(Nanomaly);
    Ytrain = X(:,[col_normal(p1(1:N_normal_train)); col_anomaly(p2(1:N_anomaly_train))]);
    Ytest = X(:,[col_normal(p1(N_normal_train+1:end)); col_anomaly(p2(N_anomaly_train+1:end))]);
    ytest = [zeros(N_normal_test,1); y(col_anomaly(p2(N_anomaly_train+1:end)))];
    N_anomaly_test_w = sum(ytest);
    
    i_param = 0;
      
    for i_over = 1:length(overcompleteness_list)
      n = round(m*overcompleteness_list(i_over));
      sigma_unif = sigma_min + (sigma_max-sigma_min) * rand(n,1); % uniform distribution of sigmas
      D0 = normc(randn(m,n));
      
      for i_s = 1 : length(lam_list)
        i_param = i_param+1;
        lam = lam_list(i_s);
        gam = gam_list(i_s);
        sigma = sigma_unif;
            
        % training phase
        % standard DL (AK-SVD)
        tic
        D1 = DL(Ytrain, D0, s_dl, nit_dl);
        t=toc;
        time_train{1}(it,i_param) = t;
        time_train{2}(it,i_param) = t;  % D1 is needed also in DL-Gauss
        % Gauss DL
        tic
        [D2, X2] = dl_gauss_l1(Ytrain, D1, sigma, lam, gam, nit_gauss_dl, nit_fista, step_fista);
        t=toc;
        time_train{2}(it,i_param) = time_train{2}(it,i_param) + t;

        % testing phase
        % AK-SVD + Gauss-OMP-L1
        tic
        errs = zeros(Ntest,1);
        probas = zeros(Ntest,1);
        objs = zeros(Ntest,1);
        for k = 1 : Ntest
          [x, Da, err, proba, obj] = gauss_l1_s(Ytest(:,k), D1, sigma, ...
                                                  lam, gam, nit_gauss_repr, nit_fista, step_fista);
          errs(k) = err(end);
          probas(k) = proba(end);
          objs(k) = obj(end);
        end
        res_err{1}(it,i_param) = mean(errs);
        [~,~,~,roc_auc] = perfcurve(ytest, errs, 1);
        res_roc_err{1}(it,i_param) = roc_auc;
        [~,~,~,roc_auc] = perfcurve(ytest, probas, 1);
        res_roc_proba{1}(it,i_param) = roc_auc;
        [~,~,~,roc_auc] = perfcurve(ytest, objs, 1);
        res_roc_obj{1}(it,i_param) = roc_auc;
        t=toc;
        time_test{1}(it,i_param) = t;

        % DL-Gauss-L1
        tic
        errs = zeros(Ntest,1);
        probas = zeros(Ntest,1);
        objs = zeros(Ntest,1);
        for k = 1 : Ntest
          [x, Da, err, proba, obj] = gauss_l1_s(Ytest(:,k), D2, sigma, ...
                                                  lam, gam, nit_gauss_repr, nit_fista, step_fista);
          errs(k) = err(end);
          probas(k) = proba(end);
          objs(k) = obj(end);
        end
        res_err{2}(it,i_param) = mean(errs);
        [~,~,~,roc_auc] = perfcurve(ytest, errs, 1);
        res_roc_err{2}(it,i_param) = roc_auc;
        [~,~,~,roc_auc] = perfcurve(ytest, probas, 1);
        res_roc_proba{2}(it,i_param) = roc_auc;
        [~,~,~,roc_auc] = perfcurve(ytest, objs, 1);
        res_roc_obj{2}(it,i_param) = roc_auc;
        t=toc;
        time_test{2}(it,i_param) = time_test{2}(it,i_param) + t;
      end
    end
  end

  for im = 1 : n_methods
    res_methods_err{im}(id,:) = mean(res_roc_err{im});
    res_methods_proba{im}(id,:) = mean(res_roc_proba{im});
    res_methods_obj{im}(id,:) = mean(res_roc_obj{im});
    time_methods_train{im}(id,:) = mean(time_train{im});
    time_methods_test{im}(id,:) = mean(time_test{im});
    fprintf("%25s[err]  : %4f\n", method_names{im}, max(res_methods_err{im}(id,:)))
    fprintf("%25s[proba]: %4f\n", method_names{im}, max(res_methods_proba{im}(id,:)))
    fprintf("%25s[obj]  : %4f\n", method_names{im}, max(res_methods_obj{im}(id,:)))
  end
  % write in file
  for im = 1 : n_methods
    fprintf(fid_res, "%s,%s,", data_file_names{id}, method_names{im});
    for ip = 1 : n_params
      fprintf(fid_res, "%f,", res_methods_err{im}(id,ip));
      fprintf(fid_res, "%f,", res_methods_proba{im}(id,ip));
      fprintf(fid_res, "%f,", res_methods_obj{im}(id,ip));
    end
    fprintf(fid_res, "\n");
  end
end

fclose(fid_res);

fprintf("\n%25s  Time train: %f\n", method_names{1}, sum(time_methods_train{1}))
fprintf("%25s  Time train: %f\n", method_names{2}, sum(time_methods_train{2}))
fprintf("%25s  Time test : %f\n", method_names{1}, sum(time_methods_test{1}))
fprintf("%25s  Time test : %f\n", method_names{2}, sum(time_methods_test{2}))

